#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "io.h"

#define BT_EOL 0
#define BT_Wall 1
#define BT_Ball 2
#define BT_Cylinder 3
#define BT_BallCC 5

typedef struct proto_ {
  int num,kind;
  double im,iI,Fx,Fy,Fz,r,mu;
} PROTO;

typedef struct part_ {
  double x,y,z,fx,fy,fz,vx,vy,vz,ox,oy,oz,nx,ny,nz,rx,ry,rz,phi;
  int num,proto;
} PART;

typedef struct cont_ {
  double dx,dy,dz;
  int i1,i2;
} CONT;

FILE *out;
int sizeofline,protonum,partnum,contnum;
PROTO *proto,*prototmp;
PART *part,*parttmp;
CONT *cont,*conttmp;

/* prototypes */
void reset_io();
int io_detfiletype(void);
void io_setfilename(char *);

void Usage(int a)
{
  switch (a) {
    case 0:
      fprintf(stderr,"usage: snp -[agn] [-o outfile] snp_files\n");
      fprintf(stderr,"\t -a: append outfile (no header will be written)\n");
      fprintf(stderr,"\t -g: group by prototype\n");
      fprintf(stderr,"\t -n: do not write header\n");
      fprintf(stderr,"\t -f: do not write forces\n");
      exit(0);
      break;
    case 1:
      fprintf(stderr,"snp: outfile exists! Remove it.\n");
      exit(0);
      break;
  }
}

void readline(FILE *in)
{
  do {
    io.getline(in);
  } while ((io.buff[0]=='#') && (!io.eof(in)));
}

int main(np,para)
char *para[];
{
  int groupproto,protoi,parti,conti,a,b,c,texture,printhead,append,dim;
  int force,argbase;
  unsigned int head[3];
  double mu,tmpmu,par[4],L,Lx,Ly,Lz,t,nx,ny,nz,sp,tnx,tny,tnz,anx,any,anz;
  char ch,*outfile;
  size_t I,D;

  sizeofline=10000;
  io.buff=(char *)malloc(sizeofline);
  protonum=10000;
  proto=(PROTO *)calloc(protonum,sizeof(PROTO));
  partnum=10000;
  part=(PART *)calloc(partnum,sizeof(PART));
  contnum=10000;
  cont=(CONT *)calloc(contnum,sizeof(CONT));
  outfile=NULL;
  groupproto=0;
  printhead=1;
  append=0;
  force=1;
  I=sizeof(int);
  D=sizeof(double);
  reset_io();
  while ((ch=getopt(np,para,"afgh-no:"))!=EOF) {
    switch (ch) {
      case 'a':
	append=1;
	break;
      case 'f':
	force=0;
	break;
      case 'g':
	groupproto=1;
	break;
      case 'h':
      case '-':
	Usage(0);
	break;
      case 'n':
	printhead=0;
	break;
      case 'o':
	outfile=strdup(optarg);
	if (!append) {
	  out=fopen(outfile,"r");
	  if (out!=NULL) {
	    fclose(out);
	    Usage(1);
	  }
	}
	break;
    }
  }
  if (optind>=np) Usage(0);
  if (outfile) {
    if (append) {
      out=fopen(outfile,"ab");
    } else {
      out=fopen(outfile,"wb");
    }
  } else {
    out=stdout;
  }
  if ((!append) && (printhead)) {
    sprintf(io.buff,"#type 6 %d %d -1\n#\n",I,D);
    fputs(io.buff,out);
  }
  argbase=optind;
  while (np>optind) {
    io_setfilename(para[optind]);
    io_detfiletype();
    io.in=io.open(io.filename,"rb");
    if (io.in) {
      fprintf(stderr,"Processing file: %d/%d\n",optind-argbase,np-argbase);
      readline(io.in);
      if (strncmp(io.buff,"name",4)==0) {
	/* 2d */
	dim=2;
	/* size */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strncmp(io.buff,"Lx",2)!=0));
	sscanf(&io.buff[5],"%lf",&L);
	Lx=L;
	readline(io.in);
	sscanf(&io.buff[5],"%lf",&Ly);
	if (Ly>L) L=Ly;
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strncmp(io.buff,"ax",2)!=0));
	sscanf(&io.buff[5],"%lf",&Lz);
	/* t */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
	    (strncmp(io.buff,"t_now",5)!=0));
	sscanf(&io.buff[8],"%lf",&t);
	/* mu */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
	    (strncmp(io.buff,"mu_default",10)!=0));
	sscanf(&io.buff[13],"%lf",&mu);
	/* prototypes */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strlen(io.buff)>1));
	readline(io.in);
	protoi=0;
	do {
	  if (protoi>=protonum) {
	    prototmp=(PROTO *)calloc(protonum*2,sizeof(PROTO));
	    memcpy(prototmp,proto,sizeof(PROTO)*protonum);
	    free(proto);
	    proto=prototmp;
	    protonum*=2;
	  }
	  sscanf(io.buff,"%d %d %d (%lf,%lf)",&proto[protoi].num,
	      &proto[protoi].kind,&a,&proto[protoi].Fx,
	      &proto[protoi].Fz);
	  proto[protoi].Fy=0.0;
	  if (proto[protoi].kind==2) {
	    fprintf(stderr,"snp: polygon is not supported yet\n");
	  }
	  if (proto[protoi].kind==4) proto[protoi].kind=2;
	  readline(io.in);
	  if ((proto[protoi].kind==1) || (proto[protoi].kind==4)) {
	    /* Sphere and Cylinder */
	    sscanf(io.buff," %lf %lf",&proto[protoi].im,
	      &proto[protoi].r);
	  } else if (proto[protoi].kind==2) {
	    /* Wall */
	    proto[protoi].r=L/2.0;
	    sscanf(io.buff," %lf %lf",&proto[protoi].im,
	      &proto[protoi].iI);
	    proto[protoi].im=1.0/proto[protoi].im;
	    proto[protoi].iI=1.0/proto[protoi].iI;
	  }
	  proto[protoi].mu=mu;
	  protoi++;
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strcmp(io.buff,"-1\n")!=0));
	/* Skip section mu */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strlen(io.buff)>1));
	/* particles */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strlen(io.buff)>1));
	readline(io.in);
	parti=0;
	do {
	  if (parti>=partnum) {
	    parttmp=(PART *)calloc(partnum*2,sizeof(PART));
	    memcpy(parttmp,part,sizeof(PART)*partnum);
	    free(part);
	    part=parttmp;
	    partnum*=2;
	  }
	  sscanf(io.buff,"%d %d (%lf,%lf) %lf (%lf,%lf) %lf\n",
		&part[parti].num,&part[parti].proto,
		&part[parti].x,&part[parti].z,&part[parti].y,
		&part[parti].vx,&part[parti].vz,&part[parti].oy);
	  part[parti].ox=part[parti].oz=0.0;
	  part[parti].rz=cos(part[parti].y);
	  part[parti].rx=sin(part[parti].y);
	  part[parti].ry=part[parti].y=part[parti].vy=0.0;
	  parti++;
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strcmp(io.buff,"-1\n")!=0));
	/* contacts */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strlen(io.buff)>1));
	readline(io.in);
	conti=0;
	while ((!io.eof(io.in)) && (strcmp(io.buff,";\n")!=0)) {
	  if (conti>=contnum) {
	    conttmp=(CONT *)calloc(contnum*2,sizeof(CONT));
	    memcpy(conttmp,cont,sizeof(CONT)*contnum);
	    free(cont);
	    cont=conttmp;
	    contnum*=2;
	  }
	  sscanf(io.buff,": %d %d",&cont[conti].i1,&cont[conti].i2);
	  readline(io.in);
	  if (io.buff[3]!='%') {
	    sscanf(io.buff,"%d %lf %lf",&a,
		&cont[conti].dx,&cont[conti].dz);
	    cont[conti].dy=0.0;
	    if ((cont[conti].dx!=0.0) || (cont[conti].dy!=0.0) ||
		(cont[conti].dz!=0.0)) conti++;
	  }
	  readline(io.in);
	}
      } else {
	/* 3d */
	dim=3;
	/* t */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
	    (strncmp(io.buff,"_t_now__(current_time)______",28)!=0));
	sscanf(&io.buff[28],"%lf",&t);
	/* mu */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
	    (strncmp(io.buff,"_mu_____(default_friction)__",28)!=0));
	sscanf(&io.buff[28],"%lf",&mu);
	/* size */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
	    (strncmp(io.buff,"_Lx__(system_x_size)________",28)!=0));
	sscanf(&io.buff[28],"%lf",&L);
	Lx=L;
	readline(io.in);
	sscanf(&io.buff[28],"%lf",&Ly);
	if (Ly>L) L=Ly;
	readline(io.in);
	sscanf(&io.buff[28],"%lf",&Lz);
	if (Lz>L) L=Lz;
	/* prototypes */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strcmp(io.buff,"____section:_prototypes__\n")!=0));
	readline(io.in);
	protoi=0;
	do {
	  if (protoi>=protonum) {
	    prototmp=(PROTO *)calloc(protonum*2,sizeof(PROTO));
	    memcpy(prototmp,proto,sizeof(PROTO)*protonum);
	    free(proto);
	    proto=prototmp;
	    protonum*=2;
	  }
	  sscanf(io.buff,"%d %d",&proto[protoi].num,&proto[protoi].kind);
	  if ((proto[protoi].kind==1) || (proto[protoi].kind==4)) {
	    /* Sphere and Cylinder */
	    sscanf(io.buff,"%d %d %lf %lf (%lf,%lf,%lf) %lf\n",&proto[protoi].num,
	      &proto[protoi].kind,&proto[protoi].im,&proto[protoi].iI,
	      &proto[protoi].Fx,&proto[protoi].Fy,&proto[protoi].Fz,
	      &proto[protoi].r);
	  } else if (proto[protoi].kind==2) {
	    /* Wall */
	    sscanf(io.buff,"%d %d %lf %lf (%lf,%lf,%lf)\n",&proto[protoi].num,
	      &proto[protoi].kind,&proto[protoi].im,&proto[protoi].iI,
	      &proto[protoi].Fx,&proto[protoi].Fy,&proto[protoi].Fz);
	    proto[protoi].r=L/2.0;
	  }
	  proto[protoi].mu=mu;
	  protoi++;
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strcmp(io.buff,"^\n")!=0));
	/* mu, may be missing */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
		 (strcmp(io.buff,"____section:_friction__\n")!=0) &&
		 (strcmp(io.buff,"____section:_particles__\n")!=0));
	if (strcmp(io.buff,"____section:_friction__\n")==0) {
	  readline(io.in);
	  while ((!io.eof(io.in)) && (strcmp(io.buff,"^\n")!=0)) {
	    sscanf(io.buff,"%d %d %lf",&a,&b,&tmpmu);
	    proto[b].mu=tmpmu;
	    readline(io.in);
	  }
	}
	/* particles */
	while ((!io.eof(io.in)) &&
	       (strcmp(io.buff,"____section:_particles__\n")!=0)) {
	  readline(io.in);
	}
	readline(io.in);
	parti=0;
	do {
	  if (parti>=partnum) {
	    parttmp=(PART *)calloc(partnum*2,sizeof(PART));
	    memcpy(parttmp,part,sizeof(PART)*partnum);
	    free(part);
	    part=parttmp;
	    partnum*=2;
	  }
	  sscanf(io.buff,"%d %d",&part[parti].num, &part[parti].proto);
	  if (proto[part[parti].proto].kind==1) {
	    /* Sphere */
	    sscanf(io.buff,"%d %d (%lf,%lf,%lf) (%lf,%lf,%lf) (%lf,%lf,%lf)\n",
		&part[parti].num,&part[parti].proto,
		&part[parti].x, &part[parti].y,&part[parti].z,
		&part[parti].vx,&part[parti].vy,&part[parti].vz,
		&part[parti].ox,&part[parti].oy,&part[parti].oz);
	  } else if ((proto[part[parti].proto].kind==2) ||
		     (proto[part[parti].proto].kind==4)) {
	    /* Wall  and Cylinder */
	    sscanf(io.buff,"%d %d (%lf,%lf,%lf) (%lf,%lf,%lf) (%lf,%lf,%lf) (%lf,%lf,%lf)\n",
		&part[parti].num,&part[parti].proto,
		&part[parti].x, &part[parti].y,&part[parti].z,
		&part[parti].nx,&part[parti].ny,&part[parti].nz,
		&part[parti].vx,&part[parti].vy,&part[parti].vz,
		&part[parti].ox,&part[parti].oy,&part[parti].oz);
	  }
	  parti++;
	  readline(io.in);
	} while ((!io.eof(io.in)) && (strcmp(io.buff,"^\n")!=0));
	/* contacts */
	do {
	  readline(io.in);
	} while ((!io.eof(io.in)) &&
	    (strcmp(io.buff,"____section:_pairs_&_contacts__\n")!=0));
	readline(io.in);
	conti=0;
	while ((!io.eof(io.in)) && (strcmp(io.buff,"^\n")!=0)) {
	  if (conti>=contnum) {
	    conttmp=(CONT *)calloc(contnum*2,sizeof(CONT));
	    memcpy(conttmp,cont,sizeof(CONT)*contnum);
	    free(cont);
	    cont=conttmp;
	    contnum*=2;
	  }
	  sscanf(io.buff,"%d %d (%lf,%lf,%lf)",&cont[conti].i1,&cont[conti].i2,
		&cont[conti].dx,&cont[conti].dy,&cont[conti].dz);
	  if ((cont[conti].dx!=0.0) || (cont[conti].dy!=0.0) ||
	      (cont[conti].dz!=0.0)) conti++;
	  readline(io.in);
	}
      }

      /* Calculate force */
      Lx/=2.0; Ly/=2.0; Lz/=2.0;
      for (a=0;a<parti;a++) {
	part[a].fx=proto[part[a].proto].Fx;
	part[a].fy=proto[part[a].proto].Fy;
	part[a].fz=proto[part[a].proto].Fz;
	part[a].x-=Lx;
	part[a].y-=Ly;
	part[a].z-=Lz;
	if ((proto[part[a].proto].kind==2) ||
	    (proto[part[a].proto].kind==4)) {
	  /* calculate orienation */
	  mu=sqrt(part[a].nx*part[a].nx+part[a].ny*part[a].ny+
						part[a].nz*part[a].nz);
	  part[a].nx/=mu; part[a].ny/=mu; part[a].nz/=mu;
	  part[a].rx=-part[a].ny;part[a].ry=part[a].nx;part[a].rz=0.0;
	  part[a].phi=asin(sqrt(part[a].nx*part[a].nx+part[a].ny*part[a].ny));
	}
      }
      for (a=0;a<conti;a++) {
	part[cont[a].i1].fx+=cont[a].dx;
	part[cont[a].i1].fy+=cont[a].dy;
	part[cont[a].i1].fz+=cont[a].dz;
	part[cont[a].i2].fx+=cont[a].dx;
	part[cont[a].i2].fy+=cont[a].dy;
	part[cont[a].i2].fz+=cont[a].dz;
      }

      par[0]=t;
      fwrite(par,D,1,out);
      texture=0;
      a=0;
      do {
	b=a;
	if (groupproto) {
	  while ((part[b].proto==part[a].proto) && (b<parti)) b++;
	} else {
	  while ((proto[part[b].proto].kind==proto[part[a].proto].kind) &&
		 (b<parti)) b++;
	}
	if (proto[part[a].proto].kind==1) {
	  /* Sphere */
	  head[0]=165669;
	  /* 5+32+256+512+1024+32768+131072=BT_BallCC+BP_Pos+BP_Velo+BP_Force+
	     BP_AngVelo+BP_Handle+BP_MatPar */
	} else if (proto[part[a].proto].kind==2) {
	  /* Wall */
	  head[0]=165729;
	  /* 1+32+64+256+512+1024+32768+131072=BT_BallCC+BP_Pos+BP_Velo+
	     BP_Force+ BP_AngVelo+BP_Handle+BP_MatPar */
	} else if (proto[part[a].proto].kind==4) {
	  /* Cylinder */
	  head[0]=165731;
	  /* 3+32+64+256+512+1024+32768+131072=BT_BallCC+BP_Pos+BP_Velo+
	     BP_Force+ BP_AngVelo+BP_Handle+BP_MatPar */
	}
	head[1]=4294967040UL+texture++;
	head[2]=b-a;
	fwrite(head,I,3,out);
	for (c=a;c<b;c++) {
	  /* diameter */
	  par[0]=proto[part[c].proto].r*2.0;
	  fwrite(par,D,1,out);
	  if (proto[part[c].proto].kind==1) {
	    head[0]=part[c].proto;
	    fwrite(head,I,1,out);
	  } else if (proto[part[c].proto].kind==2) {
	    fwrite(par,D,1,out);
	  } else if (proto[part[c].proto].kind==4) {
	    par[0]=L;
	    fwrite(par,D,1,out);
	    part[c].x-=part[c].nx*L/2.0;
	    part[c].y-=part[c].ny*L/2.0;
	    part[c].z-=part[c].nz*L/2.0;
	  }
	  /* pos */
	  par[0]=part[c].x; par[1]=part[c].y; par[2]=part[c].z;
	  fwrite(par,D,3,out);
	  if ((proto[part[a].proto].kind==2) ||
	      (proto[part[a].proto].kind==4)) {
	    /* orient */
	    par[0]=part[c].rx; par[1]=part[c].ry; par[2]=part[c].rz;
	    par[3]=part[c].phi;
	    fwrite(par,D,4,out);
	  }
	  /* velo */
	  par[0]=part[c].vx; par[1]=part[c].vy; par[2]=part[c].vz;
	  fwrite(par,D,3,out);
	  /* force */
	  par[0]=part[c].fx; par[1]=part[c].fy; par[2]=part[c].fz;
	  fwrite(par,D,3,out);
	  /* angvelo */
	  par[0]=part[c].ox; par[1]=part[c].oy; par[2]=part[c].oz;
	  fwrite(par,D,3,out);
	  /* hand */
	  head[0]=part[c].num;
	  fwrite(head,I,1,out);
	  /* matpar */
	  par[0]=proto[part[c].proto].mu;
	  if ((proto[part[c].proto].kind==1) && (dim==2)) {
	    par[1]=1.0/(M_PI*proto[part[c].proto].r
	      *proto[part[c].proto].r*proto[part[c].proto].im);
	  } else {
	    par[1]=proto[part[c].proto].im;
	  }
	  par[2]=proto[part[c].proto].iI;
	  fwrite(par,D,3,out);
	}
	a=b;
      } while (a<parti);

      if (force) {
	/* Forces */
	head[0]=614;
	head[1]=524032L;
	head[2]=conti;
	fwrite(head,I,3,out);
	for (a=0;a<conti;a++) {
	  /* pos */
	  if (proto[part[cont[a].i2].proto].kind==1) {
	    nx=part[cont[a].i2].x-part[cont[a].i1].x;
	    ny=part[cont[a].i2].y-part[cont[a].i1].y;
	    nz=part[cont[a].i2].z-part[cont[a].i1].z;
	    if (nx>Lx) nx-=Lx*2.0;
	    if (nx<-Lx) nx+=Lx*2.0;
	    if (ny>Ly) ny-=Ly*2.0;
	    if (ny<-Ly) ny+=Ly*2.0;
	    if (nz>Lz) nz-=Lz*2.0;
	    if (nz<-Lz) nz+=Lz*2.0;
	    mu=nx*nx+ny*ny+nz*nz;
	    sp=nx*cont[a].dx+ny*cont[a].dy+nz*cont[a].dz;
	    tnx=nx*sp/mu;
	    tny=ny*sp/mu;
	    tnz=nz*sp/mu;
	    par[0]=sp/sqrt(mu);
	    par[1]=sqrt(mu);
	  } else if (proto[part[cont[a].i2].proto].kind==2) {
	    nx=part[cont[a].i2].nx;
	    ny=part[cont[a].i2].ny;
	    nz=part[cont[a].i2].nz;
	    nx*=-1;
	    ny*=-1;
	    nz*=-1;
	    mu=(part[cont[a].i2].x-part[cont[a].i1].x)*nx+
	      (part[cont[a].i2].y-part[cont[a].i1].y)*ny+
	      (part[cont[a].i2].z-part[cont[a].i1].z)*nz;
	    par[1]=fabs(mu);
	    sp=nx*cont[a].dx+ny*cont[a].dy+nz*cont[a].dz;
	    tnx=nx*sp;
	    tny=ny*sp;
	    tnz=nz*sp;
	    par[0]= fabs(sp/sqrt(fabs(mu)));
	  } else if (proto[part[cont[a].i2].proto].kind==4) {
	    nx=part[cont[a].i2].nx;
	    ny=part[cont[a].i2].ny;
	    nz=part[cont[a].i2].nz;
	    mu=(part[cont[a].i2].x-part[cont[a].i1].x)*nx+
	      (part[cont[a].i2].y-part[cont[a].i1].y)*ny+
	      (part[cont[a].i2].z-part[cont[a].i1].z)*nz;
	    nx=part[cont[a].i2].x-part[cont[a].i1].x-mu*nx;
	    ny=part[cont[a].i2].y-part[cont[a].i1].y-mu*ny;
	    nz=part[cont[a].i2].z-part[cont[a].i1].z-mu*nz;
	    mu=nx*nx+ny*ny+nz*nz;
	    par[1]=proto[part[cont[a].i2].proto].r-sqrt(fabs(mu));
	    nx/=sqrt(mu);
	    ny/=sqrt(mu);
	    nz/=sqrt(mu);
	    sp=nx*cont[a].dx+ny*cont[a].dy+nz*cont[a].dz;
	    tnx=nx*sp;
	    tny=ny*sp;
	    tnz=nz*sp;
	    par[0]= fabs(sp);
	    mu=1.0;
	    nx*=-1;
	    ny*=-1;
	    nz*=-1;
	  }
	  fwrite(par,D,2,out);
	  head[0]=cont[a].i1;
	  head[1]=cont[a].i2;
	  fwrite(head,I,2,out);
	  par[0]=part[cont[a].i1].x; par[1]=part[cont[a].i1].y;
	  par[2]=part[cont[a].i1].z;
	  fwrite(par,D,3,out);
	  mu=sqrt(fabs(mu));
	  nx/=mu; ny/=mu; nz/=mu;
	  anx=ny;any=-nx;anz=0.0;
	  if ((fabs(anx)+fabs(any))==0.0) {
	    anx=1.0;par[3]=0.0;
	  } else {
	    par[3]=-asin(sqrt(anx*anx+any*any));
	  }
	  if (nz<0.0) {
	    par[3]=M_PI-par[3];
	  }
	  par[0]=anx;par[1]=any;par[2]=anz;
	  fwrite(par,D,4,out);
	  par[0]=cont[a].dx;
	  par[1]=cont[a].dy;
	  par[2]=cont[a].dz;
	  fwrite(par,D,3,out);
	}
      }

      /* EOL */
      head[0]=0;
      fwrite(head,I,1,out);
    } else {
      fprintf(stderr,"snp: input file `%s' is unreadable\n",para[optind]);
    }
    io.close(io.in);
    optind++;
  }

  /* Vege */
  if (outfile) {
    fclose(out);
  }
  return(1);
}
